@@ -37,47 +37,30 @@ impl Picker {
         self.filtered_entries.len()
     }
 
-    pub fn move_cursor(&mut self, offset: isize) {
-        if self.filtered_entries.is_empty() {
-            return;
-        }
+    pub fn clear_cursor(&mut self) {
+        self.cursor = None;
+    }
 
-        let end_index = self.filtered_entries.len() - 1;
-        let cursor = match self.cursor {
-            Some(ref mut cursor) => cursor,
+    pub fn move_cursor(&mut self, offset: isize) {
+        let end_index = match self.filtered_entries.len().checked_sub(1) {
+            Some(i) => i,
+            None => return,
+        };
+
+        match &mut self.cursor {
+            Some(cursor) => {
+                let mut index = *cursor as isize;
+                index = index + offset;
+                index = index.max(0);
+
+                *cursor = end_index.min(index as _);
+            }
             None => {
                 if self.len() > 0 {
                     self.cursor = Some(0);
                 }
-                return;
             }
         };
-
-        if offset > 0 {
-            let mut offset = offset as usize;
-            if *cursor == end_index {
-                offset -= 1;
-                *cursor = 0;
-            }
-
-            if offset < end_index - *cursor {
-                *cursor += offset;
-            } else {
-                *cursor = end_index;
-            }
-        } else if offset < 0 {
-            let mut offset = (-offset) as usize;
-            if *cursor == 0 {
-                offset -= 1;
-                *cursor = end_index;
-            }
-
-            if offset < *cursor {
-                *cursor -= offset;
-            } else {
-                *cursor = 0;
-            }
-        }
     }
 
     pub fn update_scroll(&mut self, max_height: usize) -> usize {
@@ -122,14 +105,22 @@ impl Picker {
         let _ = fmt::write(entry, args);
     }
 
-    pub fn add_custom_entry_filtered(&mut self, name: &str, pattern: &str) {
-        self.add_custom_entry(name);
-        if self.filter_custom_entry(self.custom_entries_len - 1, pattern) {
-            self.filtered_entries
-                .sort_unstable_by(|a, b| b.score.cmp(&a.score));
+    pub fn add_custom_filtered_entries<'picker, 'pattern>(
+        &'picker mut self,
+        pattern: &'pattern str,
+    ) -> AddCustomFilteredEntryGuard<'picker, 'pattern> {
+        AddCustomFilteredEntryGuard {
+            picker: self,
+            pattern,
+            needs_sorting: false,
         }
     }
 
+    pub fn sort_filtered_entries(&mut self) {
+        self.filtered_entries
+            .sort_unstable_by(|a, b| b.score.cmp(&a.score));
+    }
+
     pub fn filter(&mut self, word_indices: WordIndicesIter, pattern: &str) {
         self.filtered_entries.clear();
 
@@ -201,6 +192,31 @@ fn filtered_to_picker_entry<'a>(
     }
 }
 
+pub struct AddCustomFilteredEntryGuard<'picker, 'pattern> {
+    picker: &'picker mut Picker,
+    pattern: &'pattern str,
+    needs_sorting: bool,
+}
+impl<'picker, 'pattern> AddCustomFilteredEntryGuard<'picker, 'pattern> {
+    pub fn add(&mut self, name: &str) {
+        self.picker.add_custom_entry(name);
+        let matched = self
+            .picker
+            .filter_custom_entry(self.picker.custom_entries_len - 1, self.pattern);
+        self.needs_sorting = self.needs_sorting || matched;
+    }
+}
+impl<'picker, 'pattern> Drop for AddCustomFilteredEntryGuard<'picker, 'pattern> {
+    fn drop(&mut self) {
+        if self.needs_sorting {
+            self.picker
+                .filtered_entries
+                .sort_unstable_by(|a, b| b.score.cmp(&a.score));
+        }
+    }
+}
+
+const FIRST_CHAR_SCORE: u32 = 1;
 const WORD_BOUNDARY_MATCH_SCORE: u32 = 2;
 const CONSECUTIVE_MATCH_SCORE: u32 = 3;
 
@@ -233,7 +249,7 @@ impl FuzzyMatcher {
                 let mut previous_text_char = '\0';
                 for (i, text_char) in text[previous_match.rest_index as usize..].char_indices() {
                     if text_char.eq_ignore_ascii_case(&pattern_char) {
-                        let (matched, score) = if i == 0 && previous_match.rest_index != 0 {
+                        let (matched, mut score) = if i == 0 && previous_match.rest_index != 0 {
                             (true, CONSECUTIVE_MATCH_SCORE)
                         } else if !text_char.is_ascii_alphanumeric() {
                             (true, 0)
@@ -244,7 +260,12 @@ impl FuzzyMatcher {
                                     && text_char.is_ascii_uppercase());
                             (is_word_boundary, WORD_BOUNDARY_MATCH_SCORE)
                         };
+
                         if matched {
+                            if i == 0 && previous_match.rest_index == 0 {
+                                score += FIRST_CHAR_SCORE;
+                            }
+
                             let rest_index =
                                 previous_match.rest_index + (i + text_char.len_utf8()) as u32;
                             let score = previous_match.score + score;
@@ -290,33 +311,33 @@ mod tests {
         assert_eq!(0, fuzzy_matcher.score("a", "xyz"));
 
         assert_eq!(
-            WORD_BOUNDARY_MATCH_SCORE + CONSECUTIVE_MATCH_SCORE * 3 + 1,
+            FIRST_CHAR_SCORE + WORD_BOUNDARY_MATCH_SCORE + CONSECUTIVE_MATCH_SCORE * 3 + 1,
             fuzzy_matcher.score("word", "word"),
         );
 
         assert_eq!(
-            WORD_BOUNDARY_MATCH_SCORE + CONSECUTIVE_MATCH_SCORE * 2,
+            FIRST_CHAR_SCORE + WORD_BOUNDARY_MATCH_SCORE + CONSECUTIVE_MATCH_SCORE * 2,
             fuzzy_matcher.score("word", "wor"),
         );
 
         assert_eq!(0, fuzzy_matcher.score("word", "wrd"),);
 
         assert_eq!(
-            WORD_BOUNDARY_MATCH_SCORE + CONSECUTIVE_MATCH_SCORE,
+            FIRST_CHAR_SCORE + WORD_BOUNDARY_MATCH_SCORE + CONSECUTIVE_MATCH_SCORE,
             fuzzy_matcher.score("first/second", "f/s")
         );
 
         assert_eq!(
-            (WORD_BOUNDARY_MATCH_SCORE + CONSECUTIVE_MATCH_SCORE) * 2,
+            FIRST_CHAR_SCORE + (WORD_BOUNDARY_MATCH_SCORE + CONSECUTIVE_MATCH_SCORE) * 2,
             fuzzy_matcher.score("camelCase", "caca"),
         );
 
         assert_eq!(
-            WORD_BOUNDARY_MATCH_SCORE * 3,
+            FIRST_CHAR_SCORE + WORD_BOUNDARY_MATCH_SCORE * 3,
             fuzzy_matcher.score("ababAbA", "aaa")
         );
         assert_eq!(
-            WORD_BOUNDARY_MATCH_SCORE * 2,
+            FIRST_CHAR_SCORE + WORD_BOUNDARY_MATCH_SCORE * 2,
             fuzzy_matcher.score("abc cde", "ac"),
         );
         assert_eq!(WORD_BOUNDARY_MATCH_SCORE, fuzzy_matcher.score("abc x", "x"));
@@ -326,12 +347,23 @@ mod tests {
             fuzzy_matcher.score("AxxBxx Abcd", "abcd")
         );
 
+        assert_eq!(
+            FIRST_CHAR_SCORE + WORD_BOUNDARY_MATCH_SCORE,
+            fuzzy_matcher.score("abc", "a")
+        );
+        assert_eq!(
+            WORD_BOUNDARY_MATCH_SCORE,
+            fuzzy_matcher.score("xyz-abc", "a")
+        );
+
         let repetition_count = 100;
         let big_repetitive_text = "a".repeat(repetition_count);
         assert_eq!(
-            WORD_BOUNDARY_MATCH_SCORE + CONSECUTIVE_MATCH_SCORE * (repetition_count - 1) as u32 + 1,
+            FIRST_CHAR_SCORE
+                + WORD_BOUNDARY_MATCH_SCORE
+                + CONSECUTIVE_MATCH_SCORE * (repetition_count - 1) as u32
+                + 1,
             fuzzy_matcher.score(&big_repetitive_text, &big_repetitive_text),
         );
     }
 }
-
